import gi
import dbus

gi.require_version('Bamf', '3')

from gi.repository import Gio
from gi.repository import Bamf

from gnomehud.utils.fuzzy import match_replace


def format_label(parts):
  separator = u'\u0020\u0020\u00BB\u0020\u0020'
  return separator.join(parts)


def normalize_label(text):
  text = match_replace('_|\.\.\.|…', '', text)
  text = match_replace('\s+', ' ', text)

  return text.strip()


class DbusGtkMenuItem(object):

  def __init__(self, item, path=[]):
    self.path   = path
    self.action = str(item.get('action', ''))
    self.accel  = str(item.get('accel', ''))
    self.label  = normalize_label(item.get('label', ''))
    self.text   = format_label(self.path + [self.label])


class DbusGtkMenu(object):

  def __init__(self, session, window):
    self.results      = {}
    self.actions      = {}
    self.session      = session
    self.bus_name     = window.get_utf8_prop('_GTK_UNIQUE_BUS_NAME')
    self.app_path     = window.get_utf8_prop('_GTK_APPLICATION_OBJECT_PATH')
    self.win_path     = window.get_utf8_prop('_GTK_WINDOW_OBJECT_PATH')
    self.menubar_path = window.get_utf8_prop('_GTK_MENUBAR_OBJECT_PATH')
    self.appmenu_path = window.get_utf8_prop('_GTK_APP_MENU_OBJECT_PATH')

  def activate(self, selection):
    action = self.actions.get(selection, '')

    if 'app.' in action:
      self.send_action(action, 'app.', self.app_path)
    elif 'win.' in action:
      self.send_action(action, 'win.', self.win_path)
    elif 'unity.' in action:
      self.send_action(action, 'unity.', self.menubar_path)

  def send_action(self, name, prefix, path):
    object    = self.session.get_object(self.bus_name, path)
    interface = dbus.Interface(object, dbus_interface='org.gtk.Actions')

    interface.Activate(name.replace(prefix, ''), [], dict())

  def get_results(self):
    paths = [self.appmenu_path, self.menubar_path]

    for path in filter(None, paths):
      object    = self.session.get_object(self.bus_name, path)
      interface = dbus.Interface(object, dbus_interface='org.gtk.Menus')
      results   = interface.Start([x for x in range(1024)])

      for menu in results:
        self.results[(menu[0], menu[1])] = menu[2]

    self.collect_entries([0, 0])

  def collect_entries(self, menu, labels=[]):
    for menu in self.results.get((menu[0], menu[1]), []):
      if 'label' in menu:
        menu_item = DbusGtkMenuItem(menu, labels)
        menu_path = labels + [menu_item.label]

        if ':submenu' in menu:
          self.collect_entries(menu[':submenu'], menu_path)
        elif 'action' in menu:
          self.actions[menu_item.text] = menu_item.action

      elif ':section' in menu:
        self.collect_entries(menu[':section'], labels)


class DbusAppMenuItem(object):

  def __init__(self, item, path=[]):
    self.path   = path
    self.action = int(item[0])
    self.accel  = item[1].get('shortcut', '')
    self.label  = normalize_label(item[1].get('label', ''))
    self.text   = format_label(self.path + [self.label])


class DbusAppMenu(object):

  def __init__(self, session, window):
    self.actions   = {}
    self.session   = session
    self.xid       = window.get_xid()
    self.interface = self.get_interface()

  def activate(self, selection):
    action = self.actions[selection]
    self.interface.Event(action, 'clicked', 0, 0)

  def get_interface(self):
    bus_name = 'com.canonical.AppMenu.Registrar'
    bus_path = '/com/canonical/AppMenu/Registrar'

    try:
      object     = self.session.get_object(bus_name, bus_path)
      interface  = dbus.Interface(object, bus_name)
      name, path = interface.GetMenuForWindow(self.xid)
      object     = self.session.get_object(name, path)
      interface  = dbus.Interface(object, 'com.canonical.dbusmenu')

      return interface
    except dbus.exceptions.DBusException:
      return None

  def get_results(self):
    if self.interface:
      results = self.interface.GetLayout(0, -1, ['label'])
      self.collect_entries(results[1], [])

  def collect_entries(self, item=None, labels=[]):
    menu_item = DbusAppMenuItem(item, labels)
    menu_path = labels

    if bool(menu_item.label) and menu_item.label != 'Root':
      menu_path = labels + [menu_item.label]

    if len(item[2]):
      for child in item[2]:
        self.collect_entries(child, menu_path)

    elif bool(menu_item.label):
      self.actions[menu_item.text] = menu_item.action


class DbusMenu:

  def __init__(self):
    self.session = dbus.SessionBus()
    self.matcher = Bamf.Matcher.get_default()
    self.window  = self.matcher.get_active_window()
    self.appmenu = DbusAppMenu(self.session, self.window)
    self.gtkmenu = DbusGtkMenu(self.session, self.window)

  @property

  def prompt(self):
    app  = self.matcher.get_active_application()
    file = app.get_desktop_file()
    info = Gio.DesktopAppInfo.new_from_filename(file)

    return info.get_string('Name')

  @property

  def actions(self):
    self.appmenu.get_results()
    self.gtkmenu.get_results()

    actions = { **self.gtkmenu.actions, **self.appmenu.actions }
    self.handle_empty(actions)

    return actions.keys()

  def activate(self, selection):
    if selection in self.gtkmenu.actions:
      self.gtkmenu.activate(selection)

    elif selection in self.appmenu.actions:
      self.appmenu.activate(selection)

  def handle_empty(self, actions):
    if not len(actions):
      alert = 'No menu items available!'
      print('Gnome HUD: WARNING: (%s) %s' % (self.prompt, alert))



#! /usr/bin/python3

import gi

gi.require_version('Keybinder', '3.0')

from gi.repository import Keybinder, GLib

from gnomehud.command import default_hud_menu
from gnomehud.command import rofi_hud_menu


def run_keybinder(callback):
  Keybinder.init()
  Keybinder.bind('<Ctrl><Alt>space', callback)

  try:
    GLib.MainLoop().run()
  except KeyboardInterrupt:
    GLib.MainLoop().quit()


def main():
  run_keybinder(default_hud_menu)


def rofi():
  run_keybinder(rofi_hud_menu)


if __name__ == "__main__":
  main()



#! /usr/bin/python3

import os
import sys
import threading

from gnomehud.handlers.default import HudMenu
from gnomehud.handlers.rofi import RofiMenu


def run_command(module, function):
  args = 'python3 -c "from gnomehud.%s import %s as run; run()"'
  args = args % (module, function)

  proc = threading.Thread(target=os.system, args=[args])
  proc.start()


def run_hud_menu(menu):
  run_command('appmenu', 'main')
  run_command('keybinder', menu)


def default_hud_menu(*args):
  menu = HudMenu()
  menu.run()


def rofi_hud_menu(*args):
  menu = RofiMenu()
  menu.run()


def main():
  if sys.stdin.isatty():
    run_hud_menu('main')
  else:
    default_hud_menu()


def rofi():
  if sys.stdin.isatty():
    run_hud_menu('rofi')
  else:
    rofi_hud_menu()


if __name__ == "__main__":
  main()



import re

from fuzzysearch import find_near_matches


def match_replace(pattern, replacement, text):
  regex = re.compile(pattern, re.IGNORECASE)
  return regex.sub(replacement, text)


def normalize_string(string):
  string = match_replace('[^\w]', ' ', string)
  string = match_replace('\s+', ' ', string)

  return string.lower().strip()


def contains_words(text, words, require_all=True):
  for word in words:
    if word in text:
      if not require_all:
        return True
    elif require_all:
      return False

  return require_all


class FuzzyMatch(object):

  def __init__(self, text):
    self.score = -1
    self.query = ''
    self.value = normalize_string(text)

  def set_query(self, query):
    if bool(query) and query != self.query:
      self.query = query
      self.score = self.get_score(query)

  def get_score(self, query):
    query = normalize_string(query)
    words = query.split(' ')

    if not contains_words(self.value, words):
      return -1

    fuzzy = find_near_matches(query, self.value, max_l_dist=1)
    score = sum(map(lambda m: m.dist, fuzzy))

    return score



import dbus
import dbus.service

from gi.repository import GLib

BUS_NAME = 'com.canonical.AppMenu.Registrar'
BUS_PATH = '/com/canonical/AppMenu/Registrar'


class AppMenuService(dbus.service.Object):
  def __init__(self):
    self.window_dict = dict()

    bus_name = dbus.service.BusName(BUS_NAME, bus=dbus.SessionBus())
    dbus.service.Object.__init__(self, bus_name, BUS_PATH)

  @dbus.service.method(BUS_NAME, in_signature='uo', sender_keyword='sender')

  def RegisterWindow(self, windowId, menuObjectPath, sender):
    self.window_dict[windowId] = [dbus.String(sender), dbus.ObjectPath(menuObjectPath)]

  @dbus.service.method(BUS_NAME, in_signature='u')

  def UnregisterWindow(self, windowId):
    if windowId in self.window_dict:
      del self.window_dict[windowId]

  @dbus.service.method(BUS_NAME, in_signature='u', out_signature='so')

  def GetMenuForWindow(self, windowId):
    if windowId in self.window_dict:
      return self.window_dict[windowId]

  @dbus.service.method(BUS_NAME)

  def GetMenus(self):
    return self.window_dict

  @dbus.service.method(BUS_NAME)

  def Q(self):
    GLib.MainLoop().quit()



import setuptools

with open('README.md', 'r') as fh:
  long_description = fh.read()

setuptools.setup(
  name='gnome-hud',
  version='1.3.1',
  author='Jonian Guveli',
  author_email='jonian@hardpixel.eu',
  description='Unity like HUD menu for the GNOME Desktop Environment',
  long_description=long_description,
  long_description_content_type='text/markdown',
  url='https://github.com/hardpixel/gnome-hud',
  packages=setuptools.find_packages(),
  data_files=[
    ('share/applications', ['gnomehud.desktop'])
  ],
  install_requires=[
    'PyGObject>=3.30.0',
    'fuzzysearch>=0.5.0'
  ],
  classifiers=[
    'Programming Language :: Python :: 3',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: POSIX :: Linux'
  ],
  project_urls={
    'Bug Reports': 'https://github.com/hardpixel/gnome-hud/issues',
    'Source': 'https://github.com/hardpixel/gnome-hud',
  },
  entry_points={
    'console_scripts': [
      'gnomehud = gnomehud.command:main',
      'gnomehud-rofi = gnomehud.command:rofi',
      'gnomehud-service = gnomehud.appmenu:main'
    ]
  }
)



import gi

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk

from subprocess import Popen
from subprocess import PIPE

from gnomehud.utils.menu import DbusMenu


def rgba_to_hex(color):
  red   = int(color.red * 255)
  green = int(color.green * 255)
  blue  = int(color.blue * 255)

  return "#{0:02x}{1:02x}{2:02x}".format(red, green, blue)


class RofiMenu:

  def __init__(self):
    self.settings  = Gtk.Settings.get_default()
    self.context   = Gtk.StyleContext()
    self.dbus_menu = DbusMenu()

    self.settings.set_property('gtk-application-prefer-dark-theme', True)

  @property

  def selection(self):
    selection = self.menu_proc.communicate()[0].decode('utf8').rstrip()
    self.menu_proc.stdin.close()

    return selection

  @property

  def items(self):
    items = self.dbus_menu.actions
    return '\n'.join(items).encode('utf-8')

  @property

  def font_name(self):
    return self.settings.get_property('gtk-font-name')

  @property

  def gtk_theme_colors(self):
    colors = {
      'header':      self.lookup_color('insensitive_bg_color'),
      'base':        self.lookup_color('theme_base_color'),
      'text':        self.lookup_color('theme_text_color'),
      'bg':          self.lookup_color('theme_bg_color'),
      'borders':     self.lookup_color('borders'),
      'disabled':    self.lookup_color('insensitive_fg_color'),
      'selected_bg': self.lookup_color('theme_selected_bg_color'),
      'selected_fg': self.lookup_color('theme_selected_fg_color'),
      'error_bg':    self.lookup_color('error_bg_color'),
      'error_fg':    self.lookup_color('error_fg_color')
    }

    for name, color in colors.items():
      colors[name] = rgba_to_hex(color)

    return colors

  @property

  def theme_colors(self):
    colors = {
      'window': [
        self.gtk_theme_colors['header'],
        self.gtk_theme_colors['borders'],
        self.gtk_theme_colors['borders']
      ],
      'normal': [
        self.gtk_theme_colors['base'],
        self.gtk_theme_colors['text'],
        self.gtk_theme_colors['base'],
        self.gtk_theme_colors['selected_bg'],
        self.gtk_theme_colors['selected_fg']
      ],
      'urgent': [
        self.gtk_theme_colors['base'],
        self.gtk_theme_colors['text'],
        self.gtk_theme_colors['base'],
        self.gtk_theme_colors['error_bg'],
        self.gtk_theme_colors['error_fg']
      ]
    }

    for name, color in colors.items():
      colors[name] = ', '.join(color)

    return colors

  @property

  def theme_string(self):
    style = """
      #window { location: north; anchor: north; border: 1px;
        width: 750px; padding: 0; margin: 32px 0 0; }

      #mainbox { spacing: 0; children: [inputbar, %s]; }

      #message { border: 1px 0 0; spacing: 0; padding: 12px;
        background-color: %s; }

      #listview { border: 1px 0 0; spacing: 0; scrollbar: false;
        padding: 0; lines: 6; background-color: @normal-background; }

      #textbox { text-color: %s; }
      #textbox { text-color: %s; }
      #inputbar { padding: 14px 12px; }
      #element { border: 0; padding: 8px 12px; }
      #textbox-prompt-colon { str: ""; }
    """

    layout = 'listview'  if self.dbus_menu.actions else 'message'
    bcolor = self.gtk_theme_colors['bg']
    tcolor = self.gtk_theme_colors['disabled']

    return style % (layout, bcolor, tcolor, tcolor)

  def lookup_color(self, key):
    return self.context.lookup_color(key)[1]

  def open_menu(self):
    options = [
      'rofi',
      '-i',
      '-dmenu',
      '-theme-str', self.theme_string,
      '-p', '⚙',
      '-mesg', 'No menu actions available!',
      '-font', self.font_name,
      '-color-window', self.theme_colors['window'],
      '-color-normal', self.theme_colors['normal'],
      '-color-urgent', self.theme_colors['urgent']
    ]

    self.menu_proc = Popen(options, stdout=PIPE, stdin=PIPE)
    self.menu_proc.stdin.write(self.items)

  def run(self):
    self.open_menu()
    self.dbus_menu.activate(self.selection)



#! /usr/bin/python3

import dbus

from gi.repository import GLib
from dbus.mainloop.glib import DBusGMainLoop

from gnomehud.utils.service import BUS_NAME
from gnomehud.utils.service import AppMenuService


def run_service():
  AppMenuService()

  try:
    GLib.MainLoop().run()
  except KeyboardInterrupt:
    GLib.MainLoop().quit()


def main():
  DBusGMainLoop(set_as_default=True)
  session_bus = dbus.SessionBus()

  if not session_bus.name_has_owner(BUS_NAME):
    run_service()


if __name__ == "__main__":
  main()



import gi
import time

gi.require_version('Gtk', '3.0')
gi.require_version('Gdk', '3.0')

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import GLib
from gi.repository import GObject

from gnomehud.utils.menu import DbusMenu
from gnomehud.utils.fuzzy import FuzzyMatch
from gnomehud.utils.fuzzy import normalize_string
from gnomehud.utils.fuzzy import match_replace


def normalize_markup(text):
  return text.replace('&', '&amp;')


def run_generator(function):
  priority  = GLib.PRIORITY_LOW
  generator = function()

  GLib.idle_add(lambda: next(generator, False), priority=priority)


def inject_custom_style(widget, style_string):
  provider = Gtk.CssProvider()
  provider.load_from_data(style_string.encode())

  screen   = Gdk.Screen.get_default()
  priority = Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
  Gtk.StyleContext.add_provider_for_screen(screen, provider, priority)


def add_style_class(widget, class_names):
  context = widget.get_style_context()
  context.add_class(class_names)


class CommandListItem(Gtk.ListBoxRow):

  value = GObject.Property(type=str)
  index = GObject.Property(type=int)
  query = GObject.Property(type=str)

  def __init__(self, *args, **kwargs):
    super(Gtk.ListBoxRow, self).__init__(*args, **kwargs)

    self.set_can_focus(False)

    self.query = self.get_property('query')
    self.value = self.get_property('value')
    self.index = self.get_property('index')
    self.fuzzy = FuzzyMatch(text=self.value)

    self.label = Gtk.Label(margin=6, margin_left=10, margin_right=10)
    self.label.set_justify(Gtk.Justification.LEFT)
    self.label.set_halign(Gtk.Align.START)

    self.connect('notify::query', self.on_query_notify)

    self.add(self.label)
    self.set_label(self.value)

    self.show_all()

  def get_label(self):
    return self.label.get_label()

  def set_label(self, text):
    self.label.set_label(normalize_markup(text))

  def set_markup(self, markup):
    self.label.set_markup(normalize_markup(markup))

  def position(self):
    return self.fuzzy.score if bool(self.query) else -1

  def visibility(self):
    return self.fuzzy.score > -1 if bool(self.query) else True

  def highlight_match(self, match):
    return '<u><b>%s</b></u>' % match.group(0)

  def highlight_matches(self):
    words = self.query.replace(' ', '|')
    value = match_replace(words, self.highlight_match, self.value)

    self.set_markup(value)

  def do_label_markup(self):
    if bool(self.query):
      self.highlight_matches()

    elif '<u>' in self.get_label():
      self.set_label(self.value)

  def on_query_notify(self, *args):
    self.fuzzy.set_query(self.query)

    if self.visibility():
      GLib.idle_add(self.do_label_markup, priority=GLib.PRIORITY_HIGH_IDLE)


class CommandList(Gtk.ListBox):

  menu_actions = GObject.Property(type=object)

  def __init__(self, *args, **kwargs):
    super(Gtk.ListBox, self).__init__(*args, **kwargs)

    self.menu_actions = self.get_property('menu-actions')
    self.select_value = ''
    self.filter_value = ''
    self.visible_rows = []
    self.selected_row = 0
    self.selected_obj = None

    self.set_sort_func(self.sort_function)
    self.set_filter_func(self.filter_function)

    self.connect('row-selected', self.on_row_selected)
    self.connect('notify::menu-actions', self.on_menu_actions_notify)

  def set_filter_value(self, value=None):
    self.visible_rows = []
    self.filter_value = normalize_string(value)

    GLib.idle_add(self.invalidate_filter_value, priority=GLib.PRIORITY_LOW)

  def invalidate_filter_value(self):
    self.invalidate_filter()

    GLib.idle_add(self.invalidate_sort, priority=GLib.PRIORITY_HIGH)
    GLib.idle_add(self.invalidate_selection, priority=GLib.PRIORITY_LOW)

  def invalidate_selection(self):
    if bool(self.filter_value):
      self.visible_rows = []
      self.foreach(self.append_visible_row)
    else:
      self.visible_rows = self.get_children()

    self.select_row_by_index(0)

  def reset_selection_state(self, index):
    if index == 0:
      self.invalidate_selection()
      return True

  def append_visible_row(self, row):
    if row.visibility():
      self.visible_rows.append(row)
      return True

  def select_row_by_index(self, index):
    if index in range(0, len(self.visible_rows)):
      self.selected_row = index
      self.selected_obj = self.visible_rows[index]

      self.selected_obj.activate()

  def get_last_row_index(self):
    return len(self.visible_rows) - 1

  def select_prev_row(self):
    lastrow = self.get_last_row_index()
    prevrow = self.selected_row - 1
    prevrow = lastrow if prevrow < 0 else prevrow

    self.select_row_by_index(prevrow)

  def select_next_row(self):
    lastrow = self.get_last_row_index()
    nextrow = self.selected_row + 1
    nextrow = 0 if nextrow > lastrow else nextrow

    self.select_row_by_index(nextrow)

  def sort_function(self, row1, row2):
    score_diff = row1.position() - row2.position()
    index_diff = row1.index - row2.index

    return score_diff or index_diff

  def filter_function(self, item):
    item.set_property('query', self.filter_value)
    return item.visibility()

  def do_list_item(self, value, index):
    command = CommandListItem(value=value, index=index)

    self.append_visible_row(command)
    self.add(command)

  def do_list_items(self):
    for index, value in enumerate(self.menu_actions):
      self.do_list_item(value, index)
      self.reset_selection_state(index)
      yield True

  def on_row_selected(self, listbox, item):
    self.select_value = item.value if item else ''

  def on_menu_actions_notify(self, *args):
    self.visible_rows = []
    self.foreach(lambda item: item.destroy())

    run_generator(self.do_list_items)


class CommandWindow(Gtk.ApplicationWindow):

  def __init__(self, *args, **kwargs):
    kwargs['type'] = Gtk.WindowType.POPUP
    super(Gtk.ApplicationWindow, self).__init__(*args, **kwargs)

    self.set_size_request(750, -1)
    self.set_keep_above(True)
    self.set_resizable(False)

    self.set_type_hint(Gdk.WindowTypeHint.UTILITY)
    self.set_position(Gtk.WindowPosition.CENTER)
    self.set_custom_position()

    self.set_skip_pager_hint(True)
    self.set_skip_taskbar_hint(True)
    self.set_destroy_with_parent(True)

    self.empty_label = Gtk.Label(margin=12)
    self.empty_label.set_label('No menu actions available!')

    self.empty_box = Gtk.Box(sensitive=False)
    self.empty_box.set_size_request(750, -1)
    self.empty_box.add(self.empty_label)

    self.command_list = CommandList()
    self.command_list.invalidate_selection()

    self.search_entry = Gtk.SearchEntry(hexpand=True, margin=2)
    self.search_entry.connect('search-changed', self.on_search_entry_changed)
    self.search_entry.set_has_frame(False)

    self.scrolled_window = Gtk.ScrolledWindow(hadjustment=None, vadjustment=None)
    self.scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
    self.scrolled_window.set_size_request(750, 210)
    self.scrolled_window.add(self.command_list)

    self.header_bar = Gtk.HeaderBar(spacing=0)
    self.header_bar.set_custom_title(self.search_entry)

    self.main_box = Gtk.Box()
    self.main_box.add(self.empty_box)
    self.main_box.add(self.scrolled_window)

    self.set_titlebar(self.header_bar)
    self.add(self.main_box)

    self.set_dark_variation()
    self.set_custom_styles()

    Gdk.event_handler_set(self.on_gdk_event)

    self.connect('show', self.on_window_show)
    self.connect('button-press-event', self.on_button_press_event)

  def set_menu_actions(self, actions):
    if actions:
      self.empty_box.hide()
      self.scrolled_window.show()

      self.command_list.set_property('menu-actions', actions)
    else:
      self.scrolled_window.hide()
      self.empty_box.show()

  def set_custom_position(self):
    position = self.get_position()
    self.move(position.root_x, 32)

  def set_dark_variation(self):
    settings = Gtk.Settings.get_default()
    settings.set_property('gtk-application-prefer-dark-theme', True)

  def set_custom_styles(self):
    styles = """entry.search.flat { border: 0; outline: 0;
      border-image: none; box-shadow: none; }

      headerbar { box-shadow: none; background: @insensitive_bg_color;
        border-radius: 0; border-width: 0 0 1px 0; }

      scrolledwindow overshoot, scrolledwindow undershoot {
        background: none; box-shadow: none; }

      scrollbar { opacity: 0; }

      window decoration { box-shadow: none; border-color: @borders;
        border-style: solid; border-width: 1px; border-radius: 0; }
    """

    inject_custom_style(self, styles)

  def grab_keyboard(self, window, status, tstamp):
    while Gdk.keyboard_grab(window, True, tstamp) != status:
      time.sleep(0.1)

  def grab_pointer(self, window, status, tstamp):
    mask = Gdk.EventMask.BUTTON_PRESS_MASK

    while Gdk.pointer_grab(window, True, mask, window, None, tstamp) != status:
      time.sleep(0.1)

  def emulate_focus_out_event(self):
    tstamp = Gdk.CURRENT_TIME
    Gdk.keyboard_ungrab(tstamp)
    Gdk.pointer_ungrab(tstamp)

    fevent = Gdk.Event(Gdk.EventType.FOCUS_CHANGE)
    self.emit('focus-out-event', fevent)

  def clicked_inside(self, event):
    size    = self.get_size()
    x_range = range(0, size.width)
    y_range = range(0, size.height)

    return int(event.x) in x_range and int(event.y) in y_range

  def on_gdk_event(self, event):
    Gtk.main_do_event(event)

  def on_window_show(self, window):
    window = self.get_window()
    status = Gdk.GrabStatus.SUCCESS
    tstamp = Gdk.CURRENT_TIME

    self.grab_keyboard(window, status, tstamp)
    self.grab_pointer(window, status, tstamp)

    self.search_entry.grab_focus()

  def on_button_press_event(self, widget, event):
    win_type = event.get_window().get_window_type()
    tmp_type = Gdk.WindowType.TEMP

    if win_type == tmp_type and not self.clicked_inside(event):
      self.emulate_focus_out_event()
      return True

  def on_search_entry_changed(self, *args):
    search_value = self.search_entry.get_text()

    self.scrolled_window.unset_placement()
    self.command_list.set_filter_value(search_value)


class HudMenu(Gtk.Application):

  def __init__(self, *args, **kwargs):
    kwargs['application_id'] = 'org.hardpixel.gnomeHUD'
    super(Gtk.Application, self).__init__(*args, **kwargs)

    self.dbus_menu = DbusMenu()

    self.set_accels_for_action('app.start', ['<Ctrl><Alt>space'])
    self.set_accels_for_action('app.quit', ['Escape'])
    self.set_accels_for_action('app.prev', ['Up'])
    self.set_accels_for_action('app.next', ['Down'])
    self.set_accels_for_action('app.execute', ['Return'])

  def add_simple_action(self, name, callback):
    action = Gio.SimpleAction.new(name, None)

    action.connect('activate', callback)
    self.add_action(action)

  def do_startup(self):
    Gtk.Application.do_startup(self)

    self.add_simple_action('start', self.on_show_window)
    self.add_simple_action('quit', self.on_hide_window)
    self.add_simple_action('prev', self.on_prev_command)
    self.add_simple_action('next', self.on_next_command)
    self.add_simple_action('execute', self.on_execute_command)

  def do_activate(self):
    self.window = CommandWindow(application=self, title='Gnome HUD')
    self.window.show_all()

    self.window.set_menu_actions(self.dbus_menu.actions)
    self.window.connect('focus-out-event', self.on_hide_window)

    self.commands = self.window.command_list
    self.commands.connect_after('button-press-event', self.on_commands_click)

  def on_show_window(self, *args):
    self.window.show()

  def on_hide_window(self, *args):
    self.window.destroy()
    self.quit()

  def on_prev_command(self, *args):
    self.commands.select_prev_row()

  def on_next_command(self, *args):
    self.commands.select_next_row()

  def on_commands_click(self, widget, event):
    if event.type == Gdk.EventType._2BUTTON_PRESS:
      self.on_execute_command()

  def on_execute_command(self, *args):
    self.dbus_menu.activate(self.commands.select_value)
    self.on_hide_window()
